@@ -22,6 +22,8 @@
 #include "avar.h"
 #include "cff.h"
 #include "cmap.h"
+#include "colr.h"
+#include "cpal.h"
 #include "cvar.h"
 #include "cvt.h"
 #include "fpgm.h"
@@ -144,6 +146,11 @@ const struct {
   { OTS_TAG_STAT, false },
   { OTS_TAG_VVAR, false },
   { OTS_TAG_CFF2, false },
+  // Color font tables.
+  // We need to parse CPAL before COLR so that the number of palette entries
+  // is known; and these tables follow fvar because COLR may use variations.
+  { OTS_TAG_CPAL, false },
+  { OTS_TAG_COLR, false },
   // We need to parse GDEF table in advance of parsing GSUB/GPOS tables
   // because they could refer GDEF table.
   { OTS_TAG_GDEF, false },
@@ -164,6 +171,19 @@ const struct {
   { 0, false },
 };
 
+bool ValidateVersionTag(ots::Font *font) {
+  switch (font->version) {
+    case 0x000010000:
+    case OTS_TAG('O','T','T','O'):
+      return true;
+    case OTS_TAG('t','r','u','e'):
+      font->version = 0x000010000;
+      return true;
+    default:
+      return false;
+  }
+}
+
 bool ProcessGeneric(ots::FontFile *header,
                     ots::Font *font,
                     uint32_t signature,
@@ -190,9 +210,8 @@ bool ProcessTTF(ots::FontFile *header,
   if (!file.ReadU32(&font->version)) {
     return OTS_FAILURE_MSG_HDR("error reading sfntVersion");
   }
-  if (!ots::IsValidVersionTag(font->version)) {
-    OTS_WARNING_MSG_HDR("invalid sfntVersion: %d", font->version);
-    font->version = 0x000010000;
+  if (!ValidateVersionTag(font)) {
+    return OTS_FAILURE_MSG_HDR("invalid sfntVersion: %d", font->version);
   }
 
   if (!file.ReadU16(&font->num_tables) ||
@@ -217,13 +236,13 @@ bool ProcessTTF(ots::FontFile *header,
   // Don't call ots_failure() here since ~25% of fonts (250+ fonts) in
   // http://www.princexml.com/fonts/ have unexpected search_range value.
   if (font->search_range != expected_search_range) {
-    OTS_WARNING_MSG_HDR("bad search range");
+    OTS_WARNING_MSG_HDR("bad table directory searchRange");
     font->search_range = expected_search_range;  // Fix the value.
   }
 
   // entry_selector is Log2(maximum power of 2 <= numTables)
   if (font->entry_selector != max_pow2) {
-    OTS_WARNING_MSG_HDR("incorrect entrySelector for table directory");
+    OTS_WARNING_MSG_HDR("bad table directory entrySelector");
     font->entry_selector = max_pow2;  // Fix the value.
   }
 
@@ -233,7 +252,7 @@ bool ProcessTTF(ots::FontFile *header,
   const uint16_t expected_range_shift =
       16 * font->num_tables - font->search_range;
   if (font->range_shift != expected_range_shift) {
-    OTS_WARNING_MSG_HDR("bad range shift");
+    OTS_WARNING_MSG_HDR("bad table directory rangeShift");
     font->range_shift = expected_range_shift;  // the same as above.
   }
 
@@ -366,9 +385,8 @@ bool ProcessWOFF(ots::FontFile *header,
   if (!file.ReadU32(&font->version)) {
     return OTS_FAILURE_MSG_HDR("error reading sfntVersion");
   }
-  if (!ots::IsValidVersionTag(font->version)) {
-    OTS_WARNING_MSG_HDR("invalid sfntVersion: %d", font->version);
-    font->version = 0x000010000;
+  if (!ValidateVersionTag(font)) {
+    return OTS_FAILURE_MSG_HDR("invalid sfntVersion: %d", font->version);
   }
 
   uint32_t reported_length;
@@ -520,6 +538,10 @@ bool ProcessWOFF2(ots::FontFile *header,
                                OTS_MAX_DECOMPRESSED_FILE_SIZE / (1024.0 * 1024.0));
   }
 
+  if (decompressed_size > output->size()) {
+    return OTS_FAILURE_MSG_HDR("Size of decompressed WOFF 2.0 font exceeds output size (%gMB)", output->size() / (1024.0 * 1024.0));
+  }
+
   std::string buf(decompressed_size, 0);
   woff2::WOFF2StringOut out(&buf);
   if (!woff2::ConvertWOFF2ToTTF(data, length, &out)) {
@@ -661,6 +683,10 @@ bool ProcessGeneric(ots::FontFile *header,
                                OTS_MAX_DECOMPRESSED_FILE_SIZE / (1024.0 * 1024.0));        
   }
 
+  if (uncompressed_sum > output->size()) {
+    return OTS_FAILURE_MSG_HDR("decompressed sum exceeds output size (%gMB)", output->size() / (1024.0 * 1024.0));
+  }
+
   // check that the tables are not overlapping.
   std::vector<std::pair<uint32_t, uint8_t> > overlap_checker;
   for (unsigned i = 0; i < font->num_tables; ++i) {
@@ -899,6 +925,8 @@ bool Font::ParseTable(const TableEntry& table_entry, const uint8_t* data,
       case OTS_TAG_CFF:  table = new OpenTypeCFF(this,  tag); break;
       case OTS_TAG_CFF2: table = new OpenTypeCFF2(this, tag); break;
       case OTS_TAG_CMAP: table = new OpenTypeCMAP(this, tag); break;
+      case OTS_TAG_COLR: table = new OpenTypeCOLR(this, tag); break;
+      case OTS_TAG_CPAL: table = new OpenTypeCPAL(this, tag); break;
       case OTS_TAG_CVAR: table = new OpenTypeCVAR(this, tag); break;
       case OTS_TAG_CVT:  table = new OpenTypeCVT(this,  tag); break;
       case OTS_TAG_FPGM: table = new OpenTypeFPGM(this, tag); break;
@@ -1084,12 +1112,6 @@ bool TablePassthru::Serialize(OTSStream *out) {
   return true;
 }
 
-bool IsValidVersionTag(uint32_t tag) {
-  return tag == 0x000010000 ||
-         // OpenType fonts with CFF data have 'OTTO' tag.
-         tag == OTS_TAG('O','T','T','O');
-}
-
 bool OTSContext::Process(OTSStream *output,
                          const uint8_t *data,
                          size_t length,
